-- HazLoot library for classifying loot according to role and restriction

-- TODO: Case sensitive roles are not ideal. Fix.
-- TODO: Trim roles and restrictions on input. Be nice to users.

HazLootList = {}
HLL = HazLootList
if not LibStub then -- not in WoW
  require("TheLoot")
  HLL.standalone = true
end

HLL.ItemCount = 0
HLL.RoleByItem = {} -- Database of items
-- Why aren't these bitmaps? Because bitmaps aren't native to Lua, this is
-- just as (in)efficient as using the bitmap libraries, and probably easier
-- for laypeople to modify.
HLL.UNCLASSIFIED = 0
HLL.ROLES = { "Random roll",  -- 100
                      "Tier piece",   -- 200
                      "Caster DPS",   -- 300
                      "Healer",       -- 400
                      "Physical DPS", -- 500
                      "Tank",         -- 600
}
HLL.ROLES[0] = "<unclassified>"
HLL.ROLEVALUE = { ["<unclassified>"]=0,
                          ["Random roll"]=100,
                          ["Tier piece"]=200,
                          ["Caster DPS"]=300,
                          ["Healer"]=400,
                          ["Physical DPS"]=500,
                          ["Tank"]=600,
}
HLL.ANYRANDOM = HLL.ROLES["Random roll"]
HLL.RESTRICTIONS = { -- "Any" is a special case: 0
                         "Death Knight", --1
                         "Druid",        --2
                         "Hunter",       --3
                         "Mage",         --4
                         "Paladin",      --5
                         "Priest",       --6
                         "Rogue",        --7
                         "Shaman",       --8
                         "Warlock",      --9
                         "Warrior",      --10
}
HLL.RESTVALUES = {
                         ["<unrestricted>"]=0,
                         ["Death Knight"]=1,
                         ["Druid"]=2,
                         ["Hunter"]=3,
                         ["Mage"]=4,
                         ["Paladin"]=5,
                         ["Priest"]=6,
                         ["Rogue"]=7,
                         ["Shaman"]=8,
                         ["Warlock"]=9,
                         ["Warrior"]=10,
}
HLL.RESTRICTIONS[0] = "<unrestricted>"

--- Add an item to the loot list
-- Parses an item role string into the item name and permitted roles. Roles in the string must
--  match the predefined rolls (TODO: customize, and integrate with HSK). Note that adding an
--  itemRoleString for an item that already exists will slinetly overwrite the original item.
-- @params itemRoleString An string of the form: itemName [ | role[: restriction] [, role[: restriction] ] ]
-- @returns the number of items now in the list, if successful, -1 if the role was invalid, or the parse failed
function HLL:AddItem(itemRoleString)
  if not itemRoleString or "" == itemRoleString then return -1 end
  local sep = string.find(itemRoleString,"|")
  if nil == sep then return HLL:AddParsedItem(itemRoleString) end
  local itemName  = string.sub(itemRoleString, 1, sep - 1)
  if string.len(itemRoleString) == sep then return HLL:AddParsedItem(itemName) end
  local roleString = string.sub(itemRoleString, sep + 1 )
  local roles = {}
  local roleValue, restValue
  for roleName in string.gmatch(roleString,"([^,]+),?") do
    if not roleName then return -1 end
    -- is there a restriction?
    sep = string.find(roleName,":")
    if nil == sep then -- No
      roleValue = HLL.ROLEVALUE[roleName] 
      restValue = 0
      if not roleValue then return -1 end
    else -- yes, what is it?
      roleValue = HLL.ROLEVALUE[string.sub(roleName, 1, sep - 1)]
      restValue = HLL.RESTVALUES[string.sub(roleName, sep + 1 )]
      if not roleValue or not restValue then return -1 end
    end
    table.insert(roles, roleValue+restValue)
  end
  return HLL:AddParsedItem(itemName, roles)
end

--- Add a name and role table to the item list
-- @params name  Item name. Must not be nil, must be a string.
-- @params roles Role table. May be nil or empty, in which case the "Any Random" role is applied.
-- @returns The new count of items, or -1 on bad input
function HLL:AddParsedItem(name, roles)
  if not name or "" == name then return -1 end
  if not roles or {} == roles then
    roles = {HLL.ANYRANDOM}
  end
  if "table" ~= type(roles) and "string" ~= type(name) then return -1 end
  local increment = (nil == HLL.RoleByItem[name] and 1 or 0)
  HLL.RoleByItem[name] = roles
  HLL.ItemCount = HLL.ItemCount + increment
  return HLL.ItemCount
end

-- Add an "Unclassified" entry for this item
function HLL:AddUnclassifiedtem(itemName)
  HLL.RoleByItem[itemName] = {HLL.UNCLASSIFIED}
  HLL.ItemCount = HLL.ItemCount + 1
end

-- Lazily parse the item string for the role table
function HLL:GetRoleTable(itemName)
  if not HLL.RoleByItem[itemName] then
    local lootString = TheLoot:GetLootString(itemName)
    if lootString then
      HLL:AddItem(lootString)
    else
      HLL:AddUnclassifiedtem(itemName)
    end
  end
  return HLL.RoleByItem[itemName]
end

-- TODO: convert to itemId for localization
function HLL:GetRolesOf(itemName)
  local roleTable = HLL.RoleByItem[itemName]
  if not roleTable then roleTable = HLL:GetRoleTable(itemName) end
  local outString = ""
  if nil == roleTable then
    return HLL.ROLES[HLL.UNCLASSIFIED]
  end
  for i,role in ipairs(roleTable) do
    if i > 1 then outString = outString .. ", " end
    itemRole, itemRestriction = HLL:GetRoleName(role)
    outString = outString .. itemRole .. (nil == itemRestriction and "" or ":"..itemRestriction)
  end
  return outString
end

function HLL:IsTierPiece(itemId)
  local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType,
        itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(itemId)
  if nil == HLL.RoleByItem[itemName] then return false end
  local role = HLL.RoleByItem[itemName][1]
  return HLL.ROLEVALUE["Tier piece"] == (role - role%100)
  -- return HLL.ROLEVALUE["Tier piece"] == (role - math.mod(role,100))
end

function HLL:GetRoleName(role)
  local restriction = role%100
  -- local restriction = math.mod(role,100)
  if 0 == restriction then return HLL.ROLES[role/100]
  elseif restriction > (#HLL.RESTRICTIONS) then return HLL.RESTRICTIONS[0]
  end
  return HLL.ROLES[(role - restriction)/100], HLL.RESTRICTIONS[restriction]
end

function HLL:HowMuchLoot()
  return HLL.ItemCount
end

-- Sort Loot: return true if item1 should be sorted before item2
-- item : {itemId,itemName,itemLink,loot_slot}
function HLL:LootOrderComparator(item1,item2)
  if nil == item2 then return true end
  local value1 = HLL:GetLootValue(item1.itemId)
  local value2 = HLL:GetLootValue(item2.itemId)
  return (value1 == value2) and (item1.itemId < item2.itemId) or (value1 < value2)
end

function HLL:GetLootValue(itemId)
  -- 1. Tier pieces
  if HLL:IsTierPiece(itemId) then
    return 1 end
  local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType,
        itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(itemId)
  -- 2. Trinkets, Rings, Necks
  if itemEquipLoc == INVTYPE_NECK
     or itemEquipLoc == INVTYPE_FINGER
     or itemEquipLoc == INVTYPE_TRINKET
  then 
    return 2 end
  -- 3. Weapons, Shields, Wands, Relics, Idols
  if itemEquipLoc == INVTYPE_WEAPON
    or itemEquipLoc == INVTYPE_SHIELD
    or itemEquipLoc == INVTYPE_2HWEAPON
    or itemEquipLoc == INVTYPE_WEAPONMAINHAND
    or itemEquipLoc == INVTYPE_WEAPONOFFHAND
    or itemEquipLoc == INVTYPE_HOLDABLE
    or itemEquipLoc == INVTYPE_RANGED
    or itemEquipLoc == INVTYPE_THROWN
    or itemEquipLoc == INVTYPE_RANGEDRIGHT
    or itemEquipLoc == INVTYPE_RELIC
    or itemSubType  == "Miscellaneous"
  then
    return 3 end
  -- TODO: figure out how to localize this
  if itemType == "Armor" then
    if itemSubType == "Plate"
      then return 4 end
    if itemSubType == "Mail"
      then return 5 end
    if itemSubType == "Leather"
      then return 6 end
    if itemSubType == "Cloth"
      then return 7 end
  end
  return 8 -- unclassified
end

--------- UTILITY FUNCTIONS --------------

function HLL:Print(message)
  if HazLoot then
    HazLoot:Print(message)
  else
    print("[HLL]        "..message)
  end
end

function HLL:Debug(message)
  if HazLoot then
    HazLoot:Debug(message)
  else
    print("[HLL][DEBUG] "..message)
  end
end
